import 'package:ioc_container/ioc_container.dart';
import 'package:test/test.dart';

class A {
  A(this.name);
  final String name;
}

class B {
  B(this.a);
  final A a;
}

class C {
  C(this.b);
  bool disposed = false;
  void dispose() => disposed = true;
  final B b;
}

class D {
  D(this.b, this.c);
  bool disposed = false;

  void dispose() {
    disposed = true;
    c.dispose();
  }

  final B b;
  final C c;
}

class AFactory {
  final Map<String, A> _as = {};
  A get(String name) {
    final a = A(name);
    _as.putIfAbsent(name, () => a);

    return a;
  }
}

class SomeService {
  SomeService(AFactory factory)
      : a = factory.get('a'),
        aa = factory.get('aa');

  final A a;
  final A aa;
}

//ignore: long-method
void main() {
  test('Basic singleton', () {
    final a = A('a');
    final builder = IocContainerBuilder()
      ..add((i) => a)
      ..add((i) => B(a));
    final container = builder.toContainer();
    expect(container.get<B>().a, a);
  });

  test('Basic Singleton 2', () {
    final a = A('a');
    final builder = IocContainerBuilder()
      ..addSingletonService(a)
      ..add((i) => B(a));
    final container = builder.toContainer();
    expect(container.get<B>().a, a);
  });

  test('Without Scoping', () {
    final a = A('a');
    final builder = IocContainerBuilder()
      ..addSingletonService(a)
      ..add((i) => B(i.get<A>()))
      ..add((i) => C(i.get<B>()))
      ..add((i) => D(i.get<B>(), i.get<C>()));
    final container = builder.toContainer();
    final d = container.get<D>();
    expect(d.c.b.a, a);
    expect(d.c.b.a.name, 'a');
    expect(container.singletons.length, 1);
    expect(identical(d.c.b, d.b), false);
  });

  test('With Scoping', () {
    final a = A('a');
    final builder = IocContainerBuilder()
      ..addSingletonService(a)
      ..add((i) => B(i.get<A>()))
      ..add((i) => C(i.get<B>()))
      ..add((i) => D(i.get<B>(), i.get<C>()));
    final container = builder.toContainer();
    final scoped = container.scoped().get<D>();
    expect(scoped.c.b.a, a);
    expect(scoped.c.b.a.name, 'a');
    expect(container.singletons.length, 1);
    expect(identical(scoped.c.b, scoped.b), true);
  });

  test('With Scoping 2', () {
    final a = A('a');
    final builder = IocContainerBuilder()
      ..addSingletonService(a)
      ..add((i) => B(i.get<A>()))
      ..add((i) => C(i.get<B>()))
      ..add((i) => D(i.get<B>(), i.get<C>()));
    final container = builder.toContainer();
    final sc = container.scoped();
    final d = sc.get<D>();
    expect(d.c.b.a, a);
    expect(d.c.b.a.name, 'a');
    expect(container.singletons.length, 1);
    expect(identical(d.c.b, d.b), true);
  });

  test('With Scoping And Disposing', () {
    final a = A('a');
    final builder = IocContainerBuilder()
      ..addSingletonService(a)
      ..add((i) => B(i.get<A>()))
      ..add<C>(
        (i) => C(i.get<B>()),
        dispose: (c) => c.dispose(),
      )
      ..add<D>(
        (i) => D(i.get<B>(), i.get<C>()),
        dispose: (d) => d.dispose(),
      );
    final container = builder.toContainer();
    final scoped = container.scoped();
    final d = scoped.get<D>();
    scoped.dispose();
    expect(d.disposed, true);
    expect(d.c.disposed, true);
  });

  test('Named Key Factory', () {
    final builder = IocContainerBuilder()
      ..addSingletonService(AFactory())
      ..add((container) => SomeService(container.get<AFactory>()));

    final container = builder.toContainer();
    final someService = container.get<SomeService>();
    expect(someService.a.name, 'a');
    expect(someService.aa.name, 'aa');
  });

  test('Test Singleton', () {
    final builder = IocContainerBuilder()
      ..addSingletonService(A('a'))
      ..add((cont) => B(cont.get<A>()));
    final container = builder.toContainer();
    final a = container.get<A>();
    final b = container.get<B>();
    expect(b.a == a, true);
  });

  test('Test Singleton 2', () {
    final builder = IocContainerBuilder()..addSingleton((c) => A('a'));
    final container = builder.toContainer();
    final a = container.get<A>();
    final aa = container.get<A>();
    expect(a == aa, true);
  });

  test('Test Singleton 3', () {
    final builder = IocContainerBuilder()
      ..addSingleton((c) => B(c.get<A>()))
      ..addSingletonService(A('a'));
    final container = builder.toContainer();
    expect(container.singletons[A], container.get<A>());
    expect(container.singletons[B], container.get<B>());
    expect(container.singletons.length, 2);
  });

  test('Test Singleton 4', () {
    final builder = IocContainerBuilder()
      ..addSingleton((c) => A('a'))
      ..addSingleton((c) => B(c.get<A>()));
    final container = builder.toContainer();
    final a = container.get<A>();
    final b = container.get<B>();
    expect(a, b.a);
  });

  test('Test Can Replace', () {
    final builder = IocContainerBuilder(allowOverrides: true)
      ..addSingletonService(A('a'))
      ..addSingletonService(A('b'));
    final container = builder.toContainer();
    final a = container.get<A>();
    expect(a.name, 'b');
  });

  test('Test Cant Replace', () {
    expect(
      () => IocContainerBuilder()
        ..addSingletonService(A('a'))
        ..addSingletonService(A('b')),
      throwsException,
    );
  });

  test('Test Readme', () {
    final a = A('a');
    final builder = IocContainerBuilder()
      ..addSingletonService(a)
      ..add((i) => B(i.get<A>()))
      ..add((i) => C(i.get<B>()))
      ..add((i) => D(i.get<B>(), i.get<C>()));
    final container = builder.toContainer();
    final d = container.get<D>();
    expect(d.c.b.a, a);
    expect(d.c.b.a.name, 'a');
  });

  test('Test Immutability', () {
    final builder = IocContainerBuilder()..addSingletonService(A('a'));
    final container = builder.toContainer();

    expect(
      () => container.serviceDefinitionsByType.addAll({
        //ignore: implicit_dynamic_type
        String: ServiceDefinition(
          (c) => 'a',
          isSingleton: true,
        ),
      }),
      throwsUnsupportedError,
    );
    final container2 = builder.toContainer();

    expect(
      () => container2.serviceDefinitionsByType.addAll({
        //ignore: implicit_dynamic_type
        String: ServiceDefinition(
          (c) => 'a',
          isSingleton: true,
        ),
      }),
      throwsUnsupportedError,
    );
  });

  test('Test Lazy', () {
    final builder = IocContainerBuilder()..addSingletonService(A('a'));
    final container = builder.toContainer(isLazy: true);
    expect(container.singletons.length, 0);
    final a = container.get<A>();
    expect(container.singletons.length, 1);
    expect(container.singletons[A] == a, true);
  });

  test('Test Zealous', () {
    final builder = IocContainerBuilder()..addSingletonService(A('a'));
    final container = builder.toContainer();
    expect(container.singletons.length, 1);
    final a = container.get<A>();
    expect(container.singletons[A] == a, true);
    //Make sure the singletons are immutable
    expect(
      () => container.singletons.addAll({String: 'a'}),
      throwsUnsupportedError,
    );
  });

  test('Test Transience', () {
    final builder = IocContainerBuilder()
      ..add((c) => A('a'))
      ..add((c) => B(c.get<A>()));
    final container = builder.toContainer();
    final a = container.get<A>();
    final b = container.get<B>();
    expect(a, isNot(b.a));
  });

  test('Test Not Found', () {
    final container = IocContainerBuilder().toContainer();
    expect(
      () => container.get<A>(),
      throwsA(
        predicate(
          (exception) =>
              exception is ServiceNotFoundException<A> &&
              exception.message == 'Service A not found',
        ),
      ),
    );
  });

  test('Test Async', () async {
    final builder = IocContainerBuilder()
      ..add(
        (c) => Future<A>.delayed(
          //Simulate doing some async work
          const Duration(milliseconds: 10),
          () => A('a'),
        ),
      );
    final container = builder.toContainer();
    final a = await container.init<A>();
    expect(a, isA<A>());
  });

  test('Test Async Singleton', () async {
    final builder = IocContainerBuilder()
      ..addSingleton(
        (c) => Future<A>.delayed(
          //Simulate doing some async work
          const Duration(milliseconds: 10),
          () => A('a'),
        ),
      );
    final container = builder.toContainer();
    final a = await container.init<A>();
    expect(a, isA<A>());
  });
}
